"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "GeospatialService", {
  enumerable: true,
  get: function () {
    return _geospatial_service.default;
  }
});
Object.defineProperty(exports, "OpensearchService", {
  enumerable: true,
  get: function () {
    return _opensearch_service.default;
  }
});

var _geospatial_service = _interopRequireDefault(require("./geospatial_service"));

var _opensearch_service = _interopRequireDefault(require("./opensearch_service"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
//# sourceMappingURL=data:application/json;charset=utf-8;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbImluZGV4LnRzIl0sIm5hbWVzIjpbXSwibWFwcGluZ3MiOiI7Ozs7Ozs7Ozs7Ozs7Ozs7OztBQUtBOztBQUNBIiwic291cmNlc0NvbnRlbnQiOlsiLypcbiAqIENvcHlyaWdodCBPcGVuU2VhcmNoIENvbnRyaWJ1dG9yc1xuICogU1BEWC1MaWNlbnNlLUlkZW50aWZpZXI6IEFwYWNoZS0yLjBcbiAqL1xuXG5pbXBvcnQgR2Vvc3BhdGlhbFNlcnZpY2UgZnJvbSAnLi9nZW9zcGF0aWFsX3NlcnZpY2UnO1xuaW1wb3J0IE9wZW5zZWFyY2hTZXJ2aWNlIGZyb20gJy4vb3BlbnNlYXJjaF9zZXJ2aWNlJztcblxuZXhwb3J0IHsgR2Vvc3BhdGlhbFNlcnZpY2UsIE9wZW5zZWFyY2hTZXJ2aWNlIH07XG4iXX0=