"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "geospatial", {
  enumerable: true,
  get: function () {
    return _geospatial.default;
  }
});
Object.defineProperty(exports, "opensearch", {
  enumerable: true,
  get: function () {
    return _opensearch.default;
  }
});
Object.defineProperty(exports, "statsRoute", {
  enumerable: true,
  get: function () {
    return _stats_router.statsRoute;
  }
});

var _geospatial = _interopRequireDefault(require("./geospatial"));

var _opensearch = _interopRequireDefault(require("./opensearch"));

var _stats_router = require("./stats_router");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
//# sourceMappingURL=data:application/json;charset=utf-8;base64,eyJ2ZXJzaW9uIjozLCJzb3VyY2VzIjpbImluZGV4LnRzIl0sIm5hbWVzIjpbXSwibWFwcGluZ3MiOiI7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7OztBQUtBOztBQUNBOztBQUNBIiwic291cmNlc0NvbnRlbnQiOlsiLypcbiAqIENvcHlyaWdodCBPcGVuU2VhcmNoIENvbnRyaWJ1dG9yc1xuICogU1BEWC1MaWNlbnNlLUlkZW50aWZpZXI6IEFwYWNoZS0yLjBcbiAqL1xuXG5pbXBvcnQgZ2Vvc3BhdGlhbCBmcm9tICcuL2dlb3NwYXRpYWwnO1xuaW1wb3J0IG9wZW5zZWFyY2ggZnJvbSAnLi9vcGVuc2VhcmNoJztcbmltcG9ydCB7IHN0YXRzUm91dGUgfSBmcm9tICcuL3N0YXRzX3JvdXRlcic7XG5cbmV4cG9ydCB7IGdlb3NwYXRpYWwsIG9wZW5zZWFyY2gsIHN0YXRzUm91dGUgfTtcbiJdfQ==