(window["customImportMapDashboards_bundle_jsonpfunction"]=window["customImportMapDashboards_bundle_jsonpfunction"]||[]).push([[15],{1071:function(module,exports,__webpack_require__){"use strict";var _interopRequireDefault=__webpack_require__(0);var _typeof=__webpack_require__(8);Object.defineProperty(exports,"__esModule",{value:true});exports.default=exports.MarkdownSimple=void 0;var _react=_interopRequireWildcard(__webpack_require__(5));var _reactMarkdown=_interopRequireDefault(__webpack_require__(1065));function _getRequireWildcardCache(nodeInterop){if(typeof WeakMap!=="function")return null;var cacheBabelInterop=new WeakMap;var cacheNodeInterop=new WeakMap;return(_getRequireWildcardCache=function _getRequireWildcardCache(nodeInterop){return nodeInterop?cacheNodeInterop:cacheBabelInterop})(nodeInterop)}function _interopRequireWildcard(obj,nodeInterop){if(!nodeInterop&&obj&&obj.__esModule){return obj}if(obj===null||_typeof(obj)!=="object"&&typeof obj!=="function"){return{default:obj}}var cache=_getRequireWildcardCache(nodeInterop);if(cache&&cache.has(obj)){return cache.get(obj)}var newObj={};var hasPropertyDescriptor=Object.defineProperty&&Object.getOwnPropertyDescriptor;for(var key in obj){if(key!=="default"&&Object.prototype.hasOwnProperty.call(obj,key)){var desc=hasPropertyDescriptor?Object.getOwnPropertyDescriptor(obj,key):null;if(desc&&(desc.get||desc.set)){Object.defineProperty(newObj,key,desc)}else{newObj[key]=obj[key]}}}newObj.default=obj;if(cache){cache.set(obj,newObj)}return newObj}var markdownRenderers={root:_react.Fragment};var MarkdownSimple=function MarkdownSimple(_ref){var children=_ref.children;return _react.default.createElement(_reactMarkdown.default,{renderers:markdownRenderers},children)};exports.MarkdownSimple=MarkdownSimple;var _default=MarkdownSimple;exports.default=_default}}]);