"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ListrEventType = void 0;
/** Type of listr internal events. */
var ListrEventType;
(function (ListrEventType) {
    ListrEventType["TITLE"] = "TITLE";
    ListrEventType["STATE"] = "STATE";
    ListrEventType["ENABLED"] = "ENABLED";
    ListrEventType["SUBTASK"] = "SUBTASK";
    ListrEventType["DATA"] = "DATA";
    ListrEventType["MESSAGE"] = "MESSAGE";
})(ListrEventType = exports.ListrEventType || (exports.ListrEventType = {}));
