/**
Gracefully restore the CLI cursor on exit.

@example
```
import restoreCursor = require('restore-cursor');

restoreCursor();
```
*/
declare function restoreCursor(): void;

export = restoreCursor;
